@@ -26,10 +26,10 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.c,v 1.122 2001/09/27 04:00:55 kazu Exp $
+ * $Id: mgp.c,v 1.129 2003/06/24 05:26:33 kazu Exp $
  */
 
-static char *mgp_version = "1.09a (20010927)";
+static char *mgp_version = "1.10a (20030624)";
 
 #include "mgp.h"
 #ifdef USE_SETLOCALE
@@ -214,11 +214,14 @@ main(argc, argv)
 #endif
 	progname = argv[0];
 
+	/* secure by default.  If you need fork/exec, use -U */
+	mgp_flag |= FL_NOFORK;
+
 	if ((p = getenv("MGPWDIR")) != NULL) {
 		mgpwdir = p;
 	}
 
-#define ACCEPTOPTS	"BdvVob:c:eg:f:hlGp:qt:Q:PST:D:CORw:X:x:nF:E:"
+#define ACCEPTOPTS	"Bd:vVob:c:eg:f:hlGp:qt:Q:PSUT:D:CORw:X:x:nF:E:"
 	while ((opt = getopt(argc, argv, ACCEPTOPTS)) != -1) {
 #undef ACCEPTOPTS
 		switch (opt) {
@@ -228,6 +231,8 @@ main(argc, argv)
 
 		case 'd':
 			mgp_flag |= FL_DEMO;
+			if (isdigit(optarg[0])) demointerval = atoi(optarg);
+			else optind --;
 			break;
 
 		case 'V':
@@ -300,6 +305,9 @@ main(argc, argv)
 		case 'S':
 			mgp_flag |= FL_NOFORK;
 			break;
+		case 'U':
+			mgp_flag &= ~FL_NOFORK;
+			break;
 
 		case 'T':
 			tsfile = optarg;
@@ -334,6 +342,8 @@ main(argc, argv)
 				mgp_flag |= FL_NOVFLIB;
 			else if (strcmp(optarg, "freetype") == 0)
 				mgp_flag |= FL_NOFREETYPE;
+			else if (strcmp(optarg, "xft") == 0)
+				mgp_flag |= FL_NOXFT;
 			else {
 				fprintf(stderr, "unknown rendering engine %s\n",
 					optarg);
@@ -446,7 +456,7 @@ genhtml(start_page)
 	FILE *txt;
 	int page;
 	char *childdebug;
-	char convdb[][3][16] = {{ "jpg", "cjpeg", "djpeg" },
+	char *convdb[][3] = {{ "jpg", "cjpeg", "djpeg" },
 				{ "png", "pnmtopng", "pngtopnm" },
 				{ NULL, NULL, NULL }};
 	int inum = 0;
@@ -459,6 +469,12 @@ genhtml(start_page)
 		}
 		if (*convdb[inum] == NULL) {
 			fprintf(stderr, "unknwon image type %s.\n", htmlimage);
+			/* print out valid image types */
+			fprintf(stderr, "Valid image types: ");
+			for (inum = 0; *convdb[inum] != 0; inum++) {
+				fprintf(stderr, "%s ", *convdb[inum] );
+			}
+			fprintf(stderr,"\n");
 			cleanup(-1);
 		}
 	}
@@ -647,7 +663,7 @@ mgp_usage(name)
 	fprintf(stderr, "\t-f <font>: Specify default vflib font\n");
 #endif /*VFLIB*/
 	fprintf(stderr, "\t-g <geometry>: Set window geometry\n");
-	fprintf(stderr, "\t-h Display this help message\n");
+	fprintf(stderr, "\t-h: Display this help message\n");
 #ifdef VFLIB
 	fprintf(stderr, "\t-l: Do not use outline font on vflib\n");
 #endif /*VFLIB*/
@@ -663,13 +679,16 @@ mgp_usage(name)
 	fprintf(stderr, "\t-B: Ignore background image\n");
 	fprintf(stderr, "\t-C: Use private colormap\n");
 	fprintf(stderr, "\t-D <dir>: Generate html pages for the presentation\n");
+	fprintf(stderr, "\t-E <name>: Use this image format in html (jpg or png)\n");
 	fprintf(stderr, "\t-F<mode>,<effect>,<value>: Use forwarding caches\n");
 	fprintf(stderr, "\t-G: Page guide is on\n");
 	fprintf(stderr, "\t-O: Obey to the window manager\n");
 	fprintf(stderr, "\t-Q <quality>: Set background image quality(0-100)\n");
 	fprintf(stderr, "\t-R: Do not perform automatic reload\n");
-	fprintf(stderr, "\t-S: Do not process directives that forks process\n");
+	fprintf(stderr, "\t-S: Do not process directives that forks process (default)\n");
+	fprintf(stderr, "\t-U: Do process directives that forks process (unsecure mode)\n");
 	fprintf(stderr, "\t-T <timestampfile>: Update timestampfile on page refresh\n");
+	fprintf(stderr, "\t-P: print stderr from image conversion tools (by default it's discarded)\n");
 	fprintf(stderr, "\t-V: Be verbose\n");
 	fprintf(stderr, "\t-X <gsdevice>: ghostscript device to use\n");
 
