@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.h,v 1.128 2001/09/17 14:48:30 nishida Exp $
+ * $Id: mgp.h,v 1.134 2003/06/06 07:18:25 nishida Exp $
  */
 
 #include <stdio.h>
@@ -61,6 +61,14 @@
 #include <X11/keysym.h>
 #include <X11/cursorfont.h>
 #include <X11/Xatom.h>
+#ifdef USE_XFT2
+#include <freetype/freetype.h>
+#include <fontconfig/fontconfig.h>
+#include <X11/Xft/Xft.h>
+#ifdef HAVE_ICONV
+#include <iconv.h>
+#endif
+#endif
 #include "image/xloadimage.h"
 
 #ifndef __P
@@ -104,6 +112,7 @@
 #define DEFAULT_HGAP		0
 #define DEFAULT_VGAP		15
 #define DEFAULT_BQUALITY	100
+#define DEFAULT_OPAQUE		100	
 
 #define XLFD_HYPHEN	14
 #define FONT_FORMAT	"-*-%s-*-*-%s-*-*-*-*-*-%s"
@@ -168,6 +177,7 @@
 #define	FL_NOSTDIN	0x2000
 #define	FL_GLYPHEDGE	0x4000
 #define	FL_FRDCACHE		0x8000
+#define	FL_NOXFT		0x10000
 
 /* 	page attribute flags 	*/
 #define PGFLAG_NODEF	0x01	/* nodefault */
@@ -403,6 +413,7 @@ struct render_state {
 	int maxdescent;
 	u_int linewidth;
 	u_int brankline;
+	u_int opaque;
 	struct ctrl *xfont;
 	struct render_object *obj, *objlast;
 };
@@ -429,6 +440,9 @@ struct render_object {
 		O_IMAGE,
 #ifdef MNG
 		O_ANIM,
+#endif /* MNG */
+#ifdef USE_XFT2
+		O_XTFONT,
 #endif /* MNG */
 		O_ICON
 	} type;
@@ -464,6 +478,16 @@ struct render_object {
 			void *key;
 		} anim;
 #endif /* MNG */
+#ifdef USE_XFT2
+		struct {
+			char *data;
+			char *fontname;
+			char *registry;
+			int len;
+			int size;
+			int charset16;
+		} xftfont;
+#endif
 	} data;
 };
 
@@ -672,6 +696,7 @@ extern void ctlinsert __P((struct ctrl **, struct ctrl *));
 extern struct ctrl *ctlalloc1 __P((u_int));
 extern void ctlfree __P((struct ctrl *));
 extern struct ctrl *ctlcopy __P((struct ctrl *));
+extern struct ctrl *ctlcopy1 __P((struct ctrl *));
 extern void debug0 __P((struct ctrl *));
 extern void debug1 __P((struct ctrl *));
 
