@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: parse.c,v 1.69 1998/08/26 05:53:50 itojun Exp $
+ * $Id: parse.c,v 1.93 2003/06/14 10:22:56 nishida Exp $
  */
 
 #include "mgp.h"
@@ -34,7 +34,7 @@
 # include <fcntl.h>
 #endif
 
-static struct ctrl *parse_text __P((u_char *, u_int));
+static struct ctrl *parse_text __P((char *, u_int));
 static void read_rc __P((FILE *, char *));
 static void read_file __P((FILE *, char *, u_int *, u_int *, int));
 static void secondpass __P((void));
@@ -51,12 +51,17 @@ extern int n_errors;
 extern struct ctrl *root;
 extern char *yyfilename;
 extern int yylineno;
+/* Prevent warning if -Wall.  */
+#ifdef __GNUC__
+extern void lex_init(u_char *);
+extern int yyparse (void);
+#endif
 
 static int filterval = 0;
 
 void
 load_file(filename)
-	u_char *filename;
+	char *filename;
 {
 	FILE *fp;
 	u_int page;
@@ -96,10 +101,14 @@ cleanup_file()
 	u_int page;
 
 	for (line = 0; line < MAXLINE; line ++) {
-		if (!default_control[line])
-			continue;
-		ctlfree(default_control[line]);
-		default_control[line] = NULL;
+		if (default_control[line]) {
+			ctlfree(default_control[line]);
+			default_control[line] = NULL;
+		}
+		if (init_control[line]) {
+			ctlfree(init_control[line]);
+			init_control[line] = NULL;
+		}
 	}
 	for (page = 0; page < MAXPAGE; page ++) {
 		for (line = 0; line < MAXLINE; line ++) {
@@ -108,6 +117,7 @@ cleanup_file()
 			ctlfree(page_control[page][line]);
 			page_control[page][line] = NULL;
 		}
+		memset(&page_attribute[page], 0, sizeof(page_attribute[page]));
 	}
 	for (line = 0; line < MAXTAB + MAXSTYLE; line++) {
 		if (!tab_control[line])
@@ -125,7 +135,7 @@ cleanup_file()
 
 static struct ctrl *
 parse_text(p, page)
-	u_char *p;
+	char *p;
 	u_int page;
 {
 	struct ctrl sentinel;
@@ -155,8 +165,10 @@ read_rc(fp, filename)
 	FILE *fp;
 	char *filename;
 {
-	u_char buf[BUFSIZ];
+	char buf[BUFSIZ];
+#if 0
 	struct ctrl *cp;
+#endif
 	int lineno;
 
 	if (2 <= parse_debug)
@@ -223,6 +235,9 @@ read_rc(fp, filename)
 		case CTL_TFONT0:
 			freetypefont0 = root->ctc_value;
 			break;
+		case CTL_TMFONT0:
+			freetypemfont0 = root->ctc_value;
+			break;
 #endif
 		case CTL_NOOP:
 			/* done in grammar.y */
@@ -244,12 +259,13 @@ read_file(fp, filename, page, line, preamble)
 	u_int *line;
 	int preamble;
 {
-	u_char buf[BUFSIZ];
+	char buf[BUFSIZ];
+	char buf2[BUFSIZ];
 	struct ctrl **ch;
 	struct ctrl *cp;
 	struct ctrl *p;
 	int line_cont;
-	u_char *infilename;
+	char *infilename;
 	struct ctrl *filtermode;
 	int filterfd = -1;
 	char filtername[MAXPATHLEN];
@@ -339,7 +355,13 @@ read_file(fp, filename, page, line, preamble)
 
 		case CTL_TAB:
 		    {
-			int i = root->cti_value - 1;
+			int i = root->cti_value;
+			if (i < 0) {
+				fprintf(stderr, "%s:%d: "
+					"invalid tab index %d\n",
+					filename, lineno, root->cti_value);
+				exit(-1);
+			}
 			if (i >= MAXTAB) {	/*XXX*/
 				/* must be a string */
 				/* find a free entry */
@@ -415,7 +437,7 @@ command:
 			continue;
 		}
 	    {
-		u_char *p;
+		char *p, *q;
 
 		p = buf + strlen(buf);
 		if (buf < p && p[-1] == '\n') {
@@ -424,8 +446,13 @@ command:
 		}
 		while (buf < p && p - buf < sizeof(buf) && p[-1] == '\\') {
 			p--;
-			if (fgets(p, sizeof(buf) - (p - buf), fp) == NULL)
+			if (fgets(buf2, sizeof(buf) - (p - buf), fp) == NULL)
 				break;
+			q = buf2;
+			/* ignore blanks */
+			while (*q && isspace(*q))
+				q++;
+			strncpy(p, q, sizeof(buf) - (p - buf));
 			p += strlen(p);
 			if (buf < p && p[-1] == '\n') {
 				p--;
@@ -433,9 +460,12 @@ command:
 			}
 		}
 	    }
-
 		if (buf[0] == '#')
 			continue;
+#if 1 /* escape # by backslash */
+		if (buf[0] == 0x5c && (buf[1] == 0x5c || buf[1] == 0x23))
+			memcpy(&buf[0], &buf[1], strlen(buf));
+#endif
 		if (buf[0] == '%') {
 			/* this is directive */
 			int pb;
@@ -622,10 +652,11 @@ command:
 					ctlappend(*ch, cp);
 
 				/*
-				 * special case: %image and %line has to be
+				 * special case: %image and %bar has to be
 				 * treated as independent item
 				 */
 				if (cp->ct_op == CTL_IMAGE
+				 || cp->ct_op == CTL_ANIM
 				 || cp->ct_op == CTL_BAR) {
 					line_cont = 0;
 					*line = *line + 1;
@@ -664,11 +695,10 @@ static void
 secondpass()
 {
 	u_int page;
-	u_int l;
+	u_int l, text;
 	struct ctrl **ch;
 	struct ctrl *cp;
 	struct ctrl *cp1;
-	struct ctrl *cp2;
 
 	/*
 	 * add CTL_PAUSE to the last line of a page.
@@ -693,6 +723,7 @@ secondpass()
 	 * split GAP into VGAP and HGAP.
 	 */
 	for (page = 1; page <= maxpage; page++) {
+		text = 1;
 		for (l = 0; l <= page_attribute[page].pg_linenum; l++) {
 			for (cp = page_control[page][l];
 			     cp;
@@ -704,15 +735,32 @@ secondpass()
 					cp1->ct_next = cp->ct_next;
 					cp->ct_next = cp1->ct_next;
 				}
-			}
+
+				/*
+				 * check if this page contains only texts
+				 */
+				if (cp->ct_op == CTL_BIMAGE || cp->ct_op == CTL_BGRAD ||
+					cp->ct_op == CTL_IMAGE  || 
+#ifdef MNG
+					cp->ct_op == CTL_ANIM   || 
+#endif /* MNG */
+					(cp->ct_op == CTL_PAUSE 
+					 	&& l != page_attribute[page].pg_linenum)) {
+						text = 0; 
+					}
+				}
 		}
+		page_attribute[page].pg_text = text;
 	}
 
+
 	/* CTL_PREFIX in tab_control should be CTL_TABPREFIX. */
 	for (l = 0; l < MAXTAB + MAXSTYLE; l++) {
 		for (cp = tab_control[l]; cp; cp = cp->ct_next) {
 			if (cp->ct_op == CTL_PREFIX)
 				cp->ct_op = CTL_TABPREFIX;
+			if (cp->ct_op == CTL_PREFIXN)
+				cp->ct_op = CTL_TABPREFIXN;
 		}
 	}
 }
@@ -730,7 +778,8 @@ thirdpass()
 	struct ctrl *cp;
 	struct ctrl *cp1;
 	struct ctrl *cp2;
-	u_char *p;
+	struct ctrl *cpt;
+	char *p;
 	struct ctrl *carryover[MAXPAGE];
 
     {
@@ -743,6 +792,7 @@ thirdpass()
 	struct ctrl *tmpint[10];
 	struct ctrl *tmplong[10];
 	struct ctrl *tmpvoid[10];
+	struct ctrl *tmparea[10];
 	struct ctrl sentinel;
 	int i;
 
@@ -751,28 +801,39 @@ thirdpass()
 	memset(tmpint, 0, sizeof(tmpint));
 	memset(tmplong, 0, sizeof(tmplong));
 	memset(tmpvoid, 0, sizeof(tmpvoid));
+	memset(tmparea, 0, sizeof(tmparea));
 
 	/* default value for page 1 */
 	tmpstr[0] = ctlalloc1(CTL_PREFIX);
 	tmpstr[0]->ctc_value = strdup("");
-#ifdef VFLIB
+#if defined(VFLIB) && defined(VFONT)
 	tmpstr[1] = ctlalloc1(CTL_VFONT);
-#else
-	tmpstr[1] = ctlalloc1(CTL_NOOP);
-#endif
-#ifdef VFONT
 	tmpstr[1]->ctc_value = strdup(VFONT);		/*XXX*/
 #else
+	tmpstr[1] = ctlalloc1(CTL_NOOP);
 	tmpstr[1]->ctc_value = strdup("");
 #endif
-	tmpstr[2] = ctlalloc1(CTL_XFONT);
-	tmpstr[2]->ctc_value = strdup(DEFAULT_X_FONT);
+	tmpstr[2] = ctlalloc1(CTL_XFONT2);
+	tmpstr[2]->ctc2_value1 = strdup(DEFAULT_X_FONT);
+	tmpstr[2]->ctc2_value2 = strdup("iso8859-1");
 #ifdef FREETYPE
 	tmpstr[3] = ctlalloc1(CTL_TFONT);
+	tmpstr[3]->ctc2_value1 = strdup(freetypefont0 ? freetypefont0 :"arial.ttf");
+	tmpstr[3]->ctc2_value2 = strdup("iso8859-1");
 #else
-	tmpstr[3] = ctlalloc1(CTL_NOOP);
-#endif
+	tmpstr[3] = ctlalloc1(CTL_NOOP);	/* CTL_TFONT */
 	tmpstr[3]->ctc_value = strdup("");
+#endif
+#ifdef FREETYPE_CHARSET16
+	tmpstr[4] = ctlalloc1(CTL_TMFONT);	/* CTL_TMFONT */
+	tmpstr[4]->ctc_value = strdup("wadalab-gothic.ttf");
+#else
+	tmpstr[4] = ctlalloc1(CTL_NOOP);	/* CTL_TMFONT */
+	tmpstr[4]->ctc_value = strdup("");
+#endif
+	tmpstr[5] = ctlalloc1(CTL_XFONT2);
+	tmpstr[5]->ctc2_value1 = strdup("k14");
+	tmpstr[5]->ctc2_value2 = strdup("jisx0208.1983-*");
 	tmplong[0] = ctlalloc1(CTL_FORE);
 	get_color(DEFAULT_FORE, &tmplong[0]->ctl_value);
 	tmplong[1] = ctlalloc1(CTL_BACK);
@@ -780,14 +841,23 @@ thirdpass()
 	tmplong[2] = ctlalloc1(CTL_CCOLOR);
 	get_color(DEFAULT_FORE, &tmplong[2]->ctl_value);
 	tmpint[0] = ctlalloc1(CTL_SIZE);
-	tmpint[0]->cti_value = DEFAULT_CHARSIZE;
+	tmpint[0]->ctf_value = DEFAULT_CHARSIZE;
 	tmpint[1] = ctlalloc1(CTL_HGAP);
 	tmpint[1]->cti_value = DEFAULT_HGAP;
 	tmpint[2] = ctlalloc1(CTL_VGAP);
 	tmpint[2]->cti_value = DEFAULT_VGAP;
 	tmpint[3] = ctlalloc1(CTL_QUALITY);
 	tmpint[3]->cti_value = DEFAULT_BQUALITY;
+#ifdef XFT2
+	tmpint[4] = ctlalloc1(CTL_OPAQUE);
+	tmpint[4]->cti_value = DEFAULT_OPAQUE;
+#endif
 	tmpvoid[0] = ctlalloc1(CTL_LEFT);
+	tmparea[0] = ctlalloc1(CTL_AREA);
+	tmparea[0]->ctar_width = 100;
+	tmparea[0]->ctar_height = 100;
+	tmparea[0]->ctar_xoff = 0;
+	tmparea[0]->ctar_yoff = 0;
 
 	/* for page 1 */
 	cp = &sentinel;
@@ -795,11 +865,24 @@ thirdpass()
 		if (!tmpstr[i])
 			continue;
 		cp->ct_next = ctlalloc1(tmpstr[i]->ct_op);
-		if (tmpstr[i]->ctc_value) {
-			cp->ct_next->ctc_value =
-				strdup(tmpstr[i]->ctc_value);
-		} else
-			cp->ct_next->ctc_value = NULL;
+		if (ctl_words[tmpstr[i]->ct_op].ctl_vtype == T_STR2) {
+			if (tmpstr[i]->ctc2_value1) {
+				cp->ct_next->ctc2_value1 =
+					strdup(tmpstr[i]->ctc2_value1);
+			} else
+				cp->ct_next->ctc2_value1 = NULL;
+			if (tmpstr[i]->ctc2_value2) {
+				cp->ct_next->ctc2_value2 =
+					strdup(tmpstr[i]->ctc2_value2);
+			} else
+				cp->ct_next->ctc2_value2 = NULL;
+		} else {
+			if (tmpstr[i]->ctc_value) {
+				cp->ct_next->ctc_value =
+					strdup(tmpstr[i]->ctc_value);
+			} else
+				cp->ct_next->ctc_value = NULL;
+		}
 		cp = cp->ct_next;
 	}
 	for (i = 0; i < 10; i++) {
@@ -813,7 +896,7 @@ thirdpass()
 		if (!tmpint[i])
 			continue;
 		cp->ct_next = ctlalloc1(tmpint[i]->ct_op);
-		cp->ct_next->cti_value = tmpint[i]->cti_value;
+		cp->ct_next->ct_val = tmpint[i]->ct_val;
 		cp = cp->ct_next;
 	}
 	for (i = 0; i < 10; i++) {
@@ -822,6 +905,16 @@ thirdpass()
 		cp->ct_next = ctlalloc1(tmpvoid[i]->ct_op);
 		cp = cp->ct_next;
 	}
+	for (i = 0; i < 10; i++) {
+		if (!tmparea[i])
+			continue;
+		cp->ct_next = ctlalloc1(tmparea[i]->ct_op);
+		cp->ct_next->ctar_width = tmparea[i]->ctar_width;
+		cp->ct_next->ctar_height = tmparea[i]->ctar_height;
+		cp->ct_next->ctar_xoff = tmparea[i]->ctar_xoff;
+		cp->ct_next->ctar_yoff = tmparea[i]->ctar_yoff;
+		cp = cp->ct_next;
+	}
 	carryover[0] = sentinel.ct_next;
 
 	/*
@@ -835,8 +928,36 @@ thirdpass()
 				switch (cp->ct_op) {
 				case CTL_PREFIX: tmpstr[0] = cp; break;
 				case CTL_VFONT: tmpstr[1] = cp; break;
-				case CTL_XFONT: tmpstr[2] = cp; break;
 				case CTL_TFONT: tmpstr[3] = cp; break;
+#ifdef FREETYPE_CHARSET16
+				case CTL_TMFONT: tmpstr[4] = cp; break;
+#endif
+				case CTL_XFONT2:
+					if (strcmp(cp->ctc2_value2,
+							"iso8859-1") == 0) {
+						tmpstr[2] = cp;
+						break;
+					}
+				    {
+					struct ctrl **cpe;	/*empty cp*/
+					cpe = (struct ctrl **)NULL;
+					for (i = 5; i < 10; i++) {
+					    if (!tmpstr[i]) {
+						if (!cpe)
+						    cpe = &tmpstr[i];
+						continue;
+					    }
+					    if (strcmp(cp->ctc2_value2,
+						tmpstr[i]->ctc2_value2) == 0) {
+						    tmpstr[i] = cp;
+						    goto xfont_ok;
+					    }
+					}
+					if (cpe)
+					    *cpe = cp;
+				    xfont_ok:
+					break;
+				    }
 
 				case CTL_FORE: tmplong[0] = cp; break;
 				case CTL_BACK: tmplong[1] = cp; break;
@@ -846,11 +967,14 @@ thirdpass()
 				case CTL_HGAP: tmpint[1] = cp; break;
 				case CTL_VGAP: tmpint[2] = cp; break;
 				case CTL_QUALITY: tmpint[3] = cp; break;
+				case CTL_OPAQUE: tmpint[4] = cp; break;
 
 				case CTL_LEFT: tmpvoid[0] = cp; break;
 				case CTL_RIGHT: tmpvoid[0] = cp; break;
 				case CTL_CENTER: tmpvoid[0] = cp; break;
 				case CTL_LEFTFILL: tmpvoid[0] = cp; break;
+
+				case CTL_AREA: tmparea[0] = cp; break;
 				}
 			}
 		}
@@ -860,11 +984,24 @@ thirdpass()
 			if (!tmpstr[i])
 				continue;
 			cp->ct_next = ctlalloc1(tmpstr[i]->ct_op);
-			if (tmpstr[i]->ctc_value) {
-				cp->ct_next->ctc_value =
-					strdup(tmpstr[i]->ctc_value);
-			} else
-				cp->ct_next->ctc_value = NULL;
+			if (ctl_words[tmpstr[i]->ct_op].ctl_vtype == T_STR2) {
+				if (tmpstr[i]->ctc2_value1) {
+					cp->ct_next->ctc2_value1 =
+						strdup(tmpstr[i]->ctc2_value1);
+				} else
+					cp->ct_next->ctc2_value1 = NULL;
+				if (tmpstr[i]->ctc2_value2) {
+					cp->ct_next->ctc2_value2 =
+						strdup(tmpstr[i]->ctc2_value2);
+				} else
+					cp->ct_next->ctc2_value2 = NULL;
+			} else {
+				if (tmpstr[i]->ctc_value) {
+					cp->ct_next->ctc_value =
+						strdup(tmpstr[i]->ctc_value);
+				} else
+					cp->ct_next->ctc_value = NULL;
+			}
 			cp = cp->ct_next;
 		}
 		for (i = 0; i < 10; i++) {
@@ -878,7 +1015,7 @@ thirdpass()
 			if (!tmpint[i])
 				continue;
 			cp->ct_next = ctlalloc1(tmpint[i]->ct_op);
-			cp->ct_next->cti_value = tmpint[i]->cti_value;
+			cp->ct_next->ct_val = tmpint[i]->ct_val;
 			cp = cp->ct_next;
 		}
 		for (i = 0; i < 10; i++) {
@@ -887,6 +1024,16 @@ thirdpass()
 			cp->ct_next = ctlalloc1(tmpvoid[i]->ct_op);
 			cp = cp->ct_next;
 		}
+		for (i = 0; i < 10; i++) {
+			if (!tmparea[i])
+				continue;
+			cp->ct_next = ctlalloc1(tmparea[i]->ct_op);
+			cp->ct_next->ctar_width = tmparea[i]->ctar_width;
+			cp->ct_next->ctar_height = tmparea[i]->ctar_height;
+			cp->ct_next->ctar_xoff = tmparea[i]->ctar_xoff;
+			cp->ct_next->ctar_yoff = tmparea[i]->ctar_yoff;
+			cp = cp->ct_next;
+		}
 
 		carryover[page] = sentinel.ct_next;
 	}
@@ -934,6 +1081,9 @@ thirdpass()
 				if (cp->ct_op == CTL_TEXT
 				 || cp->ct_op == CTL_IMAGE
 				 || cp->ct_op == CTL_BAR
+#ifdef MNG
+				 || cp->ct_op == CTL_ANIM
+#endif /* MNG */
 				 || cp->ct_op == CTL_ICON) {
 					textseen++;
 					break;
@@ -965,6 +1115,9 @@ thirdpass()
 					if ((*ch)->ct_op == CTL_TEXT
 					 || (*ch)->ct_op == CTL_IMAGE
 					 || (*ch)->ct_op == CTL_BAR
+#ifdef MNG
+					 || (*ch)->ct_op == CTL_ANIM
+#endif /* MNG */
 					 || (*ch)->ct_op == CTL_ICON) {
 					    break;
 					}
@@ -996,6 +1149,9 @@ thirdpass()
 					if (cp1->ct_op == CTL_TEXT
 					 || cp1->ct_op == CTL_IMAGE
 					 || cp1->ct_op == CTL_BAR
+#ifdef MNG
+					 || cp1->ct_op == CTL_ANIM
+#endif /* MNG */
 					 || cp1->ct_op == CTL_ICON) {
 					    cp2 = cp1;
 					}
@@ -1008,13 +1164,21 @@ thirdpass()
 				}
 			}
 		}
+		if (contseen){
+			/* we have extra cont in the last line */
+			for (cp2 = page_control[page][line-1];
+				cp2->ct_next; cp2 = cp2->ct_next); 
+			cp = ctlalloc1(CTL_LINEEND);
+			if (cp) ctlinsert(&(cp2->ct_next),  cp);
+		}
 	}
     }
 
-	/* insert CTL_TAB items into CTL_LINESTART */
+	/* insert CTL_TAB */
 	for (page = 1; page <= maxpage; page++) {
 		line = page_attribute[page].pg_linenum;
 		for (l = 0; l <= line; l++) {
+			int tab_depth = 0;
 			/*
 			 * if we don't have CTL_LINESTART, we don't add
 			 * directives here.
@@ -1037,31 +1201,47 @@ thirdpass()
 				continue;
 			p = cp->ctc_value;
 			if (p && *p == '\t') {
-				int tab_depth;
-
-				tab_depth = 0;
 				p++;
+				tab_depth++;
 				while (*p == '\t') {
 					tab_depth++;
 					p++;
 				}
 				if (p) {
-					u_char *tmp;
+					char *tmp;
 
 					tmp = cp->ctc_value;
-					cp->ctc_value = strdup(p);
+					p = cp->ctc_value = strdup(p);
 					free(tmp);
 				}
+			}
 
-				if (tab_control[tab_depth]) {
-					ctlinsert(&cp1->ct_next,
-					    ctlcopy(tab_control[tab_depth]));
+#if 0
+			/* insert CTL_TAB items into CTL_LINESTART */
+			if (tab_control[tab_depth]) {
+				ctlinsert(&cp1->ct_next,
+				    ctlcopy(tab_control[tab_depth]));
+			}
+#else
+			if (tab_control[tab_depth]) {
+				ch = &page_control[page][l];
+				for (cpt = tab_control[tab_depth]; cpt; cpt = cpt->ct_next){
+					/* Thses ctrl items should be in CTL_LINESTART */
+					if (cpt->ct_op == CTL_IMAGE || cpt->ct_op == CTL_TABPREFIX ||
+							cpt->ct_op == CTL_ICON){
+						ctlinsert(&cp1->ct_next, ctlcopy1(cpt));
+						cp1 = cp1->ct_next; 
+					} else
+					/* other ctrl items should be in the head of the ctrls */
+						ctlinsert(ch, ctlcopy1(cpt));
 				}
 			}
+#endif
+
 			/* special: style escape */
 			if (p && *p == '&') {
-				u_char *p0;
-				u_char *tmp;
+				char *p0;
+				char *tmp;
 				int i;
 
 				p0 = p;
@@ -1110,6 +1290,9 @@ thirdpass()
 			for (cp = cp1; cp->ct_next; cp = cp->ct_next) {
 				if (cp->ct_next->ct_op == CTL_TEXT
 				 || cp->ct_next->ct_op == CTL_IMAGE
+#ifdef MNG
+				 || cp->ct_next->ct_op == CTL_ANIM
+#endif /* MNG */
 				 || cp->ct_next->ct_op == CTL_ICON) {
 					break;
 				}
@@ -1140,7 +1323,7 @@ thirdpass()
 					cp1 = find_font(cp->ctc_value);
 					if (!cp1) {
 						fprintf(stderr,
-"page %d line %d: font def for \"%s\" not found: ignored\n",
+			"page %d line %d: font def for \"%s\" not found: ignored\n",
 							page, l, cp->ctc_value);
 						continue;
 					}
@@ -1150,6 +1333,24 @@ thirdpass()
 			}
 		}
 	}
+
+	/*
+	 * CTL_XFONT is now obsolete, use CTL_XFONT2.
+	 */
+	for (page = 1; page <= maxpage; page++) {
+		for (l = 0; l <= page_attribute[page].pg_linenum; l++) {
+			for (cp = page_control[page][l];
+			     cp;
+			     cp = cp->ct_next) {
+				if (cp->ct_op == CTL_XFONT) {
+					p = cp->ctc_value;
+					cp->ct_op = CTL_XFONT2;
+					cp->ctc2_value1 = p;
+					cp->ctc2_value2 = strdup("iso8859-1");
+				}
+			}
+		}
+	}
 }
 
 void
@@ -1165,12 +1366,19 @@ debug0(p)
 	case T_STR:
 		fprintf(stderr, "\"%s\"", p->ctc_value);
 		break;
+	case T_STR2:
+		fprintf(stderr, "\"%s\" \"%s\"",
+			p->ctc2_value1, p->ctc2_value2);
+		break;
 	case T_INT:
 		fprintf(stderr, "%d", p->cti_value);
 		break;
 	case T_LONG:
 		fprintf(stderr, "#%lx", p->ctl_value);
 		break;
+	case T_DOUBLE:
+	        fprintf(stderr, "#%g", p->ctf_value);
+	        break;
 	case T_VOID:
 		break;
 	case T_SP:
@@ -1223,7 +1431,18 @@ debug0(p)
 		break;
 	case CTL_ICON:
 		fprintf(stderr, "type=%s color=%x siz=%d",
-			p->ctic_value, p->ctic_color, p->ctic_size);
+			p->ctic_value, (int)p->ctic_color, (int) p->ctic_size);
+		break;
+	case CTL_VALIGN:
+		fprintf(stderr, ((p->cti_value == VL_CENTER) ? "center"
+				 : ((p->cti_value == VL_TOP) ? "top"
+				    : ((p->cti_value == VL_BOTTOM) ? "bottom"
+				       : "???"))));
+		break;
+	case CTL_AREA:
+		fprintf(stderr, "xoff=%d w=%d yoff=%d h=%d",
+			p->ctar_xoff, p->ctar_width,
+			p->ctar_yoff, p->ctar_height);
 		break;
 	default:
 		fprintf(stderr, "???");
@@ -1267,7 +1486,7 @@ debug()
 
 int
 chkfile(p)
-	u_char *p;
+	char *p;
 {
 	char buf[BUFSIZ];
 
@@ -1328,7 +1547,7 @@ ctlinsert(here, this)
 
 struct ctrl *
 ctlalloc1(op)
-	u_char op;
+	u_int op;
 {
 	struct ctrl *p;
 
@@ -1383,6 +1602,19 @@ ctlcopy(this)
 	return dst0;
 }
 
+struct ctrl *
+ctlcopy1(src)
+	struct ctrl *src;
+{
+	struct ctrl *dst;
+	if (!src)
+		return NULL;
+	dst = ctlalloc1(0);
+	memcpy(dst, src, sizeof(struct ctrl));
+	dst->ct_next = NULL;
+	return dst;
+}
+
 int
 ctlcmp(a, b)
 	struct ctrl *a;
@@ -1401,10 +1633,18 @@ ctlcmp(a, b)
 	switch (ctl_words[a->ct_op].ctl_vtype) {
 	case T_STR:
 		return strcmp(a->ctc_value, b->ctc_value);
+	case T_STR2:
+		if (strcmp(a->ctc2_value1, b->ctc2_value1) == 0
+		 && strcmp(a->ctc2_value2, b->ctc2_value2) == 0) {
+			return 0;
+		} else
+			return 1;
 	case T_INT:
 		return (a->cti_value == b->cti_value) ? 0 : 1;
 	case T_LONG:
 		return (a->ctl_value == b->ctl_value) ? 0 : 1;
+	case T_DOUBLE:
+		return (a->ctf_value == b->ctf_value) ? 0 : 1;
 	case T_VOID:
 		return 0;
 	case T_SP:
@@ -1484,8 +1724,9 @@ fsearchopen(fname, mode, path)
 		if (!p)
 			break;
 		strcpy(buf, p);
-		if (p = strrchr(buf, '/'))
+		if ((p = strrchr(buf, '/'))) {
 			p[1] = '\0';
+		}
 		strcat(buf, fname);
 		fp = fopen(buf, mode);
 		if (fp)
