@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: tfont.c,v 1.42 2001/04/12 11:49:04 nishida Exp $
+ * $Id: tfont.c,v 1.43 2002/01/12 01:56:25 nishida Exp $
  */
 
 #include "mgp.h"
@@ -590,7 +590,7 @@ CharToUnicode(code, registry)
 #endif /* FREETYPE_CHARSET16 */
 
 #if 1 /* latin2-4 encoding processing */
-			if (code > 0xa0) {
+			if (code > 0xa0 && code < 256) {
 				if (registry)
 					regid = get_regid(registry) -1;
 				else {
@@ -602,7 +602,14 @@ CharToUnicode(code, registry)
 			}
 #endif
 
+/*
+ * For freetype 1.4, suggested by mituharu@math.s.chiba-u.ac.jp 
+ */ 
+#if TT_FREETYPE_MAJOR == 1 &&  TT_FREETYPE_MINOR == 4
+			return TT_Char_Index(char_map, (u_short) code);
+#else
 			return TT_Char_Index(char_map, (short) code);
+#endif
 		}
 #ifdef FREETYPE_CHARSET16
 		else {
